from .conv import CausalConv1d, TemporalConvBlock
from .dense_interpolation import DenseInterpolation
from .mlp import AutoEncoder, MLP, PositionWiseFeedForward, ResidualSequential
from .positional_encoding import PositionalEncoding
from .se_layer import SELayer1d, SELayer2d


__all__ = [
    "AutoEncoder",
    "CausalConv1d",
    "TemporalConvBlock",
    "DenseInterpolation",
    "MLP",
    "PositionalEncoding",
    "PositionWiseFeedForward",
    "ResidualSequential",
    "SELayer1d",
    "SELayer2d",
]



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import torch


__all__ = ["calculate_accuracy"]


def calculate_accuracy(inputs: torch.Tensor, targets: torch.Tensor) -> float:
    """
    A function that calculates accuracy for batch processing.
    Returns accuracy as a Python float.

    Args:
        inputs (torch.Tensor): shape == [N, n_class]
        targets (torch.Tensor): shape == [N]

    Returns: accracy (float)
    """
    with torch.no_grad():
        total = targets.shape[0]
        _, predicted = torch.max(inputs, 1)
        correct = (predicted == targets).cpu().sum().float().item()

    return correct / total



# mypy: ignore-errors
from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F


__all__ = [
    "generate_sample_indices",
    "generate_anchor_positive_input",
    "generate_negative_input",
    "positive_criterion_for_triplet_loss",
    "negative_criterion_for_triplet_loss",
    "calculate_triplet_loss",
]


def generate_sample_indices(
    n_rand_samples: int, batch_size: int, length: int
) -> Tuple[np.ndarray, int, np.ndarray, int, torch.Tensor]:
    """
    generate sample indices.

    Args:
        n_rand_samples: the number of negative samples
        batch_size: batch size
        length: length of time series.

    Returns:


    """
    if n_rand_samples > 0 and batch_size > 0 and length > 0:
        len_pos_neg: int = np.random.randint(1, length + 1)

        # anchor
        len_anchor: int = np.random.randint(len_pos_neg, length + 1)  # len of anchors
        begin_batches: np.ndarray = np.random.randint(0, length - len_anchor + 1, size=batch_size)

        begin_pos_samples: np.ndarray = np.random.randint(0, len_anchor - len_pos_neg + 1, size=batch_size)
        begin_pos: np.ndarray = begin_batches + begin_pos_samples

        end_pos: np.ndarray = begin_pos + len_pos_neg

        begin_neg_samples: torch.Tensor = torch.randint(
            0, high=length - len_pos_neg + 1, size=(n_rand_samples, batch_size)
        )

        return begin_batches, len_anchor, end_pos, len_pos_neg, begin_neg_samples

    else:
        raise ValueError("The argument must be greater than or equal to 1.")


def generate_anchor_positive_input(
    n_rand_samples: int, batch_size: int, length: int, original_data: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
    """

    Args:
        n_rand_samples: the number of negative samples
        batch_size: batch size
        length: length of time series
        original_data: training data

    Returns:
        - anchor data (torch.Tensor)
        - positive data (torch.Tensor)

    """
    begin_batches, len_anchor, end_pos, len_pos_neg, _ = generate_sample_indices(n_rand_samples, batch_size, length)

    anchor_data = torch.cat(
        [original_data[j : j + 1, :, begin_batches[j] : begin_batches[j] + len_anchor] for j in range(batch_size)]
    )

    positive_data = torch.cat(
        [original_data[j : j + 1, :, end_pos[j] - len_pos_neg : end_pos[j]] for j in range(batch_size)]
    )

    return anchor_data, positive_data


@torch.jit.script
def generate_negative_input(
    begin_neg_samples: torch.Tensor,
    len_pos_neg: int,
    batch_size: int,
    idx: int,
    train: torch.Tensor,
    samples: torch.Tensor,
) -> torch.Tensor:
    """

    Args:
        begin_neg_samples: Starting points of negative samples
        len_pos_neg: length of negative and postive samples
        batch_size: batch size
        idx: Parameter `k` on the paper.
        train: training dataset
        samples:

    Returns:
        negative_input (torch.Tensor) - [batch_size, features, len_pos_neg]

    Notes:
        When running mypy, ``error: Slice index must be an integer or None`` is detected on lines 97 and 98.
        Ignore it for now.
        Also, I didn't add ``type: ignore`` as a comment because it has a negative impact on ``TorchScript`` execution.

    """
    negative_data = torch.cat(
        [
            train[samples[idx, j] : samples[idx, j] + 1][
                :, :, begin_neg_samples[idx, j] : begin_neg_samples[idx, j] + len_pos_neg
            ]
            for j in range(batch_size)
        ]
    )

    return negative_data


@torch.jit.script
def positive_criterion_for_triplet_loss(anchor: torch.Tensor, positive: torch.Tensor) -> torch.Tensor:
    r"""

    .. math::

        positive\ loss =
            -\log\Bigl(\sigma(f(x^{ref}) ^\mathrm{T} f(x^{pos}))\Bigr)

    Args:
        anchor: :math:`f(x^{ref})` ... anchor representation
        positive: :math:`f(x^{pos})` ... positive representation

    Returns:
        positive loss (torch.Tensor)

    """
    positive_loss: torch.Tensor = -torch.mean(F.logsigmoid(torch.bmm(anchor, positive)))
    return positive_loss


@torch.jit.script
def negative_criterion_for_triplet_loss(anchor: torch.Tensor, positive: torch.Tensor) -> torch.Tensor:
    r"""

    .. math::

        negative\ loss = - \log\Bigl(\sigma(-f(x^{ref}) ^\mathrm{T}f(x_k^{neg}))\Bigl)

    Args:
        anchor: :math:`f(x^{ref})` ... anchor representation
        positive: :math:`f(x^{neg})` ... negative representation

    Returns:
        negative loss (torch.Tensor)

    """
    negative_loss = -torch.mean(F.logsigmoid(-torch.bmm(anchor, positive)))
    return negative_loss


@torch.jit.script
def calculate_triplet_loss(
    positive_loss: torch.Tensor, negative_loss: torch.Tensor, multiplicative_ration: float
) -> torch.Tensor:
    r"""

    .. math::

        Loss = positive\ loss + α × negative\ loss

    Args:
        positive_loss: output of ``positive_criterion_for_triplet_loss``
        negative_loss: output of ``negative_criterion_for_triplet_loss``
        multiplicative_ration: :math:`α`

    Returns:
        loss (torch.Tensor)

    """
    loss: torch.Tensor = positive_loss + multiplicative_ration * negative_loss

    return loss



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************
from typing import Union

import torch
import torch.nn as nn


__all__ = ["Swish", "mish", "Mish", "FReLU1d", "FReLU2d"]


class Swish(nn.Module):
    """

    Apply Swish activate function.

    """

    def __init__(self, beta: bool = False):
        super(Swish, self).__init__()
        if beta:
            self.weight: Union[nn.Parameter, float] = nn.Parameter(torch.tensor([1.0]), requires_grad=True)
        else:
            self.weight = 1.0

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply Swish function.

        Examples:
            >>> import torch
            >>> act = Swish()
            >>> x = torch.randn(2)
            >>> y = act(x)

        Args:
            x (torch.Tensor):

        Returns:
            Result of applying Swish (torch.Tensor)

        """
        out = x * torch.sigmoid(self.weight * x)
        return out


@torch.jit.script
def mish(x: torch.Tensor) -> torch.Tensor:
    """
    Apply mish function.

    Examples:
        >>> import torch
        >>> inputs = torch.randn(2)
        >>> outputs = mish(inputs)

    Args:
        x (torch.Tensor): Input Data

    Returns:
        Result of applying mish (torch.Tensor)

    """
    return x * torch.tanh(nn.functional.softplus(x))


class Mish(nn.Module):
    """
    Apply mish activate function.

    """

    def __init__(self):
        super(Mish, self).__init__()
        self.tanh: nn.Module = nn.Tanh()
        self.softplus: nn.Module = nn.Softplus()

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """
        Apply Mish to the input.

        Examples:
            >>> import torch
            >>> act = Mish()
            >>> x = torch.randn(2)
            >>> y = act(y)

        Args:
            inputs(torch.Tensor):

        Returns:
            Result of applying mish (torch.Tensor)

        """
        return inputs * self.tanh(self.softplus(inputs))


class FReLU1d(nn.Module):
    """
    Applies the Funnel Activation (FReLU) for 1d inputs such as sensor signals.

    Examples:
        >>> inputs = torch.randn(1, 3, 128)     # [N, features, seq_len]
        >>> frelu = FReLU1d(3)
        >>> outputs = frelu(inputs)

    """

    def __init__(self, in_features: int, kernel_size: int = 3, stride: int = 1, padding: int = 1):
        super(FReLU1d, self).__init__()
        self.conv = nn.Conv1d(
            in_features,
            in_features,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=in_features,
        )
        self.norm = nn.BatchNorm1d(in_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Applies the Funnel Activation (FReLU) for 1d inputs such as sensor signals.

        Args:
            x: torch.Tensor

        Returns:

        """
        out = self.conv(x)
        out = self.norm(out)
        out = torch.max(x, out)
        return out


class FReLU2d(nn.Module):
    """
    Applies the Funnel Activation (FReLU) for 2d inputs such as images.

    Examples:
        >>> inputs = torch.randn(1, 3, 128, 128)     # [N, channels, heights, widths]
        >>> frelu = FReLU2d(3)
        >>> outputs = frelu(inputs)

    """

    def __init__(self, in_features: int, kernel_size: int = 3, stride: int = 1, padding: int = 1):
        super(FReLU2d, self).__init__()
        self.conv = nn.Conv2d(
            in_features,
            in_features,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=in_features,
        )
        self.norm = nn.BatchNorm2d(in_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Applies the Funnel Activation (FReLU) for 2d inputs such as images.

        Args:
            x: torch.Tensor

        Returns:

        """
        out = self.conv(x)
        out = self.norm(out)
        out = torch.max(x, out)
        return out



__version__ = "0.8.1"
__author__ = "Hirotaka Kawashima"
__author_email__ = "khirotaka@vivaldi.net"
__license__ = "Apache-2.0"
__copyright__ = "Copyright (c) 2020, {}.".format(__author__)
__homepage__ = "https://github.com/khirotaka/enchanter"

__docs__ = "Enchanter is a library for machine learning tasks for comet.ml users."


from enchanter import addons
from enchanter import engine
from enchanter import tasks


__all__ = ["addons", "engine", "tasks"]



from typing import Callable, Optional, Union

import numpy as np
import torch
from torch.utils.data import Dataset


__all__ = ["TimeSeriesLabeledDataset", "TimeSeriesUnlabeledDataset"]


class TimeSeriesUnlabeledDataset(Dataset):
    """
    Examples:
        >>> from torch.utils.data import DataLoader
        >>> ds = TimeSeriesUnlabeledDataset(data=...)
        >>> loader = DataLoader(ds)
        >>> data = next(iter(loader))
    """

    def __init__(
        self, data: Union[torch.Tensor, np.ndarray], transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None
    ):
        self.data: torch.Tensor = data
        self.transform = transform

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, item):
        data = self.data[item]
        if self.transform:
            data = self.transform(data)

        return data


class TimeSeriesLabeledDataset(TimeSeriesUnlabeledDataset):
    """
    Examples:
        >>> from torch.utils.data import DataLoader
        >>> ds = TimeSeriesLabeledDataset(data=..., targets=...)
        >>> loader = DataLoader(ds)
        >>> data, targets = next(iter(loader))
    """

    def __init__(
        self,
        data: Union[torch.Tensor, np.ndarray],
        targets: Union[torch.Tensor, np.ndarray],
        transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
    ):
        super(TimeSeriesLabeledDataset, self).__init__(data, transform)
        self.targets = targets

    def __getitem__(self, item):
        data = super(TimeSeriesLabeledDataset, self).__getitem__(item)
        target = self.targets[item]
        return data, target



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.utils import _single
from ...utils.backend import slice_axis


__all__ = ["CausalConv1d", "TemporalConvBlock"]


class CausalConv1d(nn.Conv1d):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        """
        Causal Conv1d

        Paper: `WaveNet: A Generative Model for Raw Audio <https://arxiv.org/abs/1609.03499>`_

        Args:
            in_channels: the number of input channels
            out_channels: the number of output channels
            kernel_size: kernel size
            stride: stride
            dilation: rate of dilation
            groups: the number of groups
            bias: if true use bias (default: True)

        """
        super(CausalConv1d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding=_single((kernel_size - 1) * dilation),
            dilation=dilation,
            groups=groups,
            bias=bias,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward propagation

        Args:
            x: [N, in_channels, L]

        Returns:
            [N, out_channels, L]

        """
        out = F.conv1d(x, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)
        if self.kernel_size[0] > 0:
            out = slice_axis(out, axis=2, begin=0, end=-self.padding[0])

        return out


class TemporalConvBlock(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        kernel_size: int,
        stride: int = 1,
        dilation: int = 1,
        dropout: float = 0.5,
        activation: nn.Module = nn.ReLU(),
        final_activation: bool = False,
    ):
        r"""
        Temporal Convolutional Block

        Paper: `An Empirical Evaluation of Generic Convolutional and Recurrent Networks for Sequence Modeling \
                <https://arxiv.org/abs/1803.01271>`_

        Args:
            in_features: the number of input channels
            out_features: the number of output channels
            kernel_size: kernel size
            stride: stride
            dilation: dilation rate
            dropout: dropout rate
            activation: activation function (default: ReLU)
            final_activation: If true, apply the activation function after the residual connection
        """
        super(TemporalConvBlock, self).__init__()
        self.final_activation = final_activation
        self.conv = nn.Sequential(
            nn.utils.weight_norm(
                CausalConv1d(in_features, out_features, kernel_size, stride=stride, dilation=dilation)
            ),
            activation,
            nn.Dropout(dropout),
            nn.utils.weight_norm(
                CausalConv1d(out_features, out_features, kernel_size, stride=stride, dilation=dilation)
            ),
            activation,
            nn.Dropout(dropout),
        )
        self.downsample = nn.Conv1d(in_features, out_features, 1) if in_features != out_features else None
        self.activation = activation

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity: torch.Tensor = x
        out: torch.Tensor = self.conv(x)

        if self.downsample is not None:
            identity = self.downsample(x)

        out = out + identity
        if self.final_activation:
            out = self.activation(out)
        return out



from typing import List, Dict, Optional, Union, Tuple

import numpy as np
from comet_ml.experiment import BaseExperiment
import torch
from torch.cuda import amp
from sklearn.svm import SVC
from sklearn.model_selection import train_test_split

from enchanter.engine import modules
from enchanter.engine import BaseRunner
from enchanter.engine.typehint import ScikitModel
from enchanter.addons.criterions.ts_triplet_loss import (
    generate_anchor_positive_input,
    generate_negative_input,
    generate_sample_indices,
    positive_criterion_for_triplet_loss,
    negative_criterion_for_triplet_loss,
    calculate_triplet_loss,
)
from enchanter.callbacks import Callback, BaseLogger
from enchanter.utils.datasets import TimeSeriesUnlabeledDataset


class TimeSeriesUnsupervisedRunner(BaseRunner):
    """
    Runner for unsupervised time series representation learning.

    Unsupervised representation learning for time series uses
    the the Unsupervised Triplet Loss proposed in NeurIPS 2019.

    Paper: `Unsupervised Scalable Representation Learning for Multivariate Time Series \
        <https://papers.nips.cc/paper/8713-unsupervised-scalable-representation-learning-for-multivariate-time-series>`_


    Examples:
        >>> experiment: Union[BaseLogger, BaseExperiment] = ...
        >>> model: torch.nn.Module = ...
        >>> optimizer: torch.optim.Optimizer = ...
        >>> runner = TimeSeriesUnsupervisedRunner(model, optimizer, experiment)
        >>> runner.add_loader("train", ...)
        >>> runner.train_config(...)
        >>> runner.run()

    """

    def __init__(
        self,
        model: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        experiment: Union[BaseExperiment, BaseLogger],
        n_negative_samples: int = 1,
        negative_penalty: int = 1,
        compared_len: Optional[int] = None,
        evaluator: ScikitModel = SVC(),
        save_memory: bool = False,
        scheduler: List = None,
        callbacks: Optional[List[Callback]] = None,
    ):
        """
        Initializer

        Args:
            model: PyTorch model which outputting a fixed-length vector regardless of the length of the input series.
            optimizer: PyTorch Optimizer
            experiment: ``comet_ml.BaseExperiment`` or ``enchanter.callbacks.BaseLogger``
            n_negative_samples: Parameter K in the paper. The number of negative samples to be sampled during training.
            negative_penalty: Coefficients that control how much negative values are valued.
            compared_len: Maximum length of randomly chosen time series. (default None).
            save_memory: If True, enables to save GPU memory.
            scheduler: lr scheduler. use ``torch.optim.lr_scheduler``
            callbacks: List of callback.
        """
        super(TimeSeriesUnsupervisedRunner, self).__init__()
        self.model = model
        self.optimizer = optimizer
        self.experiment = experiment

        if compared_len is None:
            self.compared_len = np.inf
        else:
            self.compared_len = compared_len

        self.n_negative_samples = n_negative_samples
        self.negative_penalty = negative_penalty
        self.save_memory = save_memory
        self.train_ds = None
        if scheduler is None:
            self.scheduler = list()
        else:
            self.scheduler = scheduler

        self.callbacks = callbacks
        self.evaluator = evaluator
        self.evaluator_params = None

    def initialize(self) -> None:
        super(TimeSeriesUnsupervisedRunner, self).initialize()
        if isinstance(self.loaders["train"].dataset, TimeSeriesUnlabeledDataset):
            self.train_ds = self.loaders["train"].dataset  # type: ignore
        else:
            raise ValueError(
                "You should use the `enchanter.utils.datasets.TimeSeriesUnlabeledDataset` or"
                " `TimeSeriesLabeledDataset` for the dataset class."
            )

    def calculate_negative_loss_per_negative_sample(
        self,
        begin_neg: torch.Tensor,
        len_pos_neg: int,
        batch_size: int,
        negative_sample_step: int,
        anchor_representation: torch.Tensor,
        data: torch.Tensor,
    ) -> torch.Tensor:
        """
        calculate negative loss per negative sample

        Args:
            begin_neg:
            len_pos_neg:
            batch_size:
            negative_sample_step:
            anchor_representation:
            data:

        Returns:
            negative loss (torch.Tensor)

        """
        representation_size: int = anchor_representation.shape[2]
        negative_data: torch.Tensor = generate_negative_input(
            begin_neg, len_pos_neg, batch_size, negative_sample_step, self.train_ds.data, data  # type: ignore
        ).to(
            self.device
        )  # [batch_size, features, seq_len]

        negative_representation: torch.Tensor = self.model(negative_data).view(batch_size, representation_size, 1)
        negative_loss: torch.Tensor = negative_criterion_for_triplet_loss(
            anchor_representation, negative_representation
        )
        return negative_loss

    def calculate_negative_loss(self, positive_loss, anchor_representation):
        """
        calculate negative loss using all negative samples.

        Args:
            positive_loss:
            anchor_representation:

        Returns:
            loss (torch.Tensor)

        """
        train_size: int = len(self.train_ds)
        multiplicative_ration: float = self.negative_penalty / self.n_negative_samples
        batch_size: int = anchor_representation.shape[0]
        length: int = min(self.compared_len, self.train_ds.data.shape[2])
        samples: torch.Tensor = torch.tensor(
            np.random.choice(train_size, size=(self.n_negative_samples, batch_size)), dtype=torch.long
        )

        _, _, _, len_pos_neg, begin_neg_samples = generate_sample_indices(self.n_negative_samples, batch_size, length)

        for i in range(self.n_negative_samples):
            negative_loss = self.calculate_negative_loss_per_negative_sample(
                begin_neg_samples, len_pos_neg, batch_size, i, anchor_representation, samples
            )
            positive_loss = calculate_triplet_loss(positive_loss, negative_loss, multiplicative_ration)

            if self.save_memory and i != self.n_negative_samples - 1:
                positive_loss.backward(retain_graph=True)
                positive_loss = torch.tensor(0.0, device=self.device)
                torch.cuda.empty_cache()

        return positive_loss

    def train_step(self, batch) -> Dict[str, torch.Tensor]:
        if len(batch) == 2:
            x_train, _ = batch
        else:
            x_train = batch[0]
        batch_size: int = x_train.shape[0]

        length: int = min(self.compared_len, self.train_ds.data.shape[2])  # type: ignore

        anchor_data, positive_data = generate_anchor_positive_input(
            self.n_negative_samples, batch_size, length, x_train
        )

        with amp.autocast(enabled=isinstance(self.scaler, amp.GradScaler)):
            anchor_representation: torch.Tensor = self.model(anchor_data)
            positive_representation: torch.Tensor = self.model(positive_data)

            representation_size: int = anchor_representation.shape[1]

            anchor_representation = anchor_representation.view(batch_size, 1, representation_size)
            positive_representation = positive_representation.view(batch_size, representation_size, 1)

            loss = positive_criterion_for_triplet_loss(anchor_representation, positive_representation)

            if self.save_memory:
                loss.backward(retain_graph=True)
                loss = torch.tensor(0.0, device=self.device)
                del positive_representation
                torch.cuda.empty_cache()

            loss = self.calculate_negative_loss(loss, anchor_representation)

        return {"loss": loss}

    def train_end(self, outputs: List) -> Dict[str, torch.Tensor]:
        avg_loss = torch.stack([x["loss"] for x in outputs]).mean()
        return {"avg_loss": avg_loss}

    def val_step(self, batch: Tuple) -> Dict[str, torch.Tensor]:
        return self.train_step(batch)

    def val_end(self, outputs: List) -> Dict[str, torch.Tensor]:
        return self.train_end(outputs)

    def test_step(self, batch: Tuple) -> Dict[str, torch.Tensor]:
        x, y = batch
        with amp.autocast(enabled=isinstance(self.scaler, amp.GradScaler)):
            encoded = self.model(x)
        return {"encoded": encoded, "targets": y}

    def _generate_train_features(self) -> Tuple[torch.Tensor, torch.Tensor]:
        features = []
        targets = []

        loader = self.loaders["train"]
        loader = modules.tfds_to_numpy(loader) if modules.is_tfds(loader) else loader

        self.model.eval()
        with torch.no_grad(), amp.autocast(enabled=isinstance(self.scaler, amp.GradScaler)):
            for batch in loader:
                x, y = batch
                features.append(self.model(x.to(self.device)))
                targets.append(y)

        return torch.cat(features).cpu().numpy(), torch.cat(targets).cpu().numpy()

    def test_end(self, outputs: List) -> Dict[str, torch.Tensor]:
        x_train, y_train = self._generate_train_features()

        if "grid_search" in self.manager.params.keys():
            self.evaluator.set_params(**self.manager.params["grid_search"])
            if len(y_train) <= 10000:
                self.evaluator.fit(x_train, y_train)  # type: ignore
            else:
                split = train_test_split(x_train, y_train, train_size=10000, random_state=0, stratify=y_train)
                self.evaluator.fit(split[0], split[2])  # type: ignore

            # self.experiment.log_parameters(search.best_params_, prefix="grid_search_best_params")
        else:
            self.evaluator.fit(x_train, y_train)  # type: ignore

        x_test = torch.cat([output["encoded"] for output in outputs]).cpu().numpy()
        y_test = torch.cat([output["targets"] for output in outputs]).cpu().numpy()

        return {"evaluator_score": torch.tensor(self.evaluator.score(x_test, y_test))}

    def encode(self, data: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
        """
        Output encoded data. The output data has the same data type as the input.

        Args:
            data: data

        Returns:
            encoded data

        """
        if isinstance(data, np.ndarray):
            data = torch.tensor(data, device=self.device)
            ndarray = True

        elif isinstance(data, torch.Tensor):
            data = data.to(self.device)
            ndarray = False

        else:
            raise ValueError("Unexpected data type.")

        self.model.eval()
        with torch.no_grad(), amp.autocast(enabled=isinstance(self.scaler, amp.GradScaler)):
            out: torch.Tensor = self.model(data)

        if ndarray:
            out = out.cpu().numpy()  # type: ignore

        return out

    def predict(self, x: Union[torch.Tensor, np.ndarray]) -> np.ndarray:
        """
        See Also ``self.encode``

        """
        out = self.encode(x)
        if isinstance(out, torch.Tensor):
            out = out.numpy()

        return out



import io
import sys
from typing import Tuple

import torch
import netron


__all__ = ["with_netron"]


def with_netron(
    model: torch.nn.Module,
    dummy: Tuple[torch.Tensor, ...],
    backend: str = "onnx",
    open_browser: bool = True,
    port: int = 8080,
    host: str = "localhost",
):
    """
    Visualize the PyTorch graph in a browser.

    Examples:
        >>> from enchanter.addons.layers import AutoEncoder
        >>> x = torch.randn(1, 32)  # [N, in_features]
        >>> model = AutoEncoder([32, 16, 8, 2])
        >>> with_netron(model, (x, ))

    Warnings:
        Models that cannot be graphed using TorchScript or ONNX cannot be visualized.

    Args:
        model: PyTorch model
        dummy: dummy input for generate graph
        backend: specified graph format. [torchscript or onnx]. default onnx.
        open_browser: if True, open browser.
        port: port number. default: 8080.
        host: hostname. default: localhost.

    """
    if backend.lower() == "torchscript":
        extension = "pth"
    elif backend.lower() == "onnx":
        extension = "onnx"
    else:
        raise ValueError("Unknown backend: {}".format(backend))

    model_name: str = model.__class__.__name__
    buffer = io.BytesIO()

    if extension == "pth":
        model = torch.jit.trace(model, dummy)
        torch.jit.save(model, buffer)

    else:
        torch.onnx.export(model, dummy, buffer)

    netron.serve("{}.{}".format(model_name, extension), buffer.getvalue(), browse=open_browser, port=port, host=host)
    sys.stdout.write("Press CTRL+C to quit.\n")
    netron.wait()



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import List, Callable, Union
from torch import relu
from torch.tensor import Tensor
from torch.nn import Module, ModuleList, Linear, Sequential, Conv1d, ReLU


__all__ = ["MLP", "PositionWiseFeedForward", "ResidualSequential", "AutoEncoder"]


class MLP(Module):
    """
    Class to create MLP.

    Args:
        shapes (List[int]): The number of neurons in each layer of MLP. Assuming an array consisting of int type elements.
                The value of the 0th element of the given array is treated as the number of input dimensions to the model.

        activation (Union[Callable[[torch.Tensor], torch.Tensor], nn.Module]): Activation Function.
                    Differentiable Callable objects such as ``torch.relu`` and ``enchanter.addons.Mish()``

    Examples:
        >>> import enchanter.addons as addons
        >>> model = addons.layers.MLP([10, 512, 128, 5], addons.Mish())
        >>> print(model)
        >>> # ModuleList(
        >>> #    (0): Linear(in_features=10, out_features=512, bias=True)
        >>> #    (1): Linear(in_features=512, out_features=128, bias=True)
        >>> #    (2): Linear(in_features=128, out_features=5, bias=True)
        >>> #)

    """

    def __init__(self, shapes: List[int], activation: Union[Callable[[Tensor], Tensor], Module] = relu):
        super(MLP, self).__init__()
        layers: List[Module] = []
        self.activation: Callable[[Tensor], Tensor] = activation

        for i in range(len(shapes) - 1):
            layers.append(Linear(shapes[i], shapes[i + 1]))

        self.layers: Module = ModuleList(layers)

    def forward(self, x: Tensor) -> Tensor:
        """
        forward propagation processing.

        Args:
            x: input data

        Returns:

        """
        for layer in self.layers[:-1]:  # type: ignore
            x = layer(x)
            x = self.activation(x)

        x = self.layers[-1](x)  # type: ignore
        return x


class PositionWiseFeedForward(Module):
    """
    ``PositionWise FeedForward`` proposed in `Attention is all you need`. This class uses ``1x1 Conv1d`` internally.


    Args:
        d_model: the number of expected features in the Position Wise Feed Forward inputs.
        expansion: Magnification rate of the number of hidden dimensions. Default: 2

    Examples:
        >>> import torch
        >>> import enchanter.addons as addons
        >>> x = torch.randn(1, 128, 512)    # [N, seq_len, features]
        >>> ff = addons.layers.PositionWiseFeedForward(512)
        >>> out = ff(x)

    """

    def __init__(self, d_model: int, expansion: int = 2) -> None:
        super(PositionWiseFeedForward, self).__init__()
        self.conv: Module = Sequential(
            Conv1d(d_model, d_model * expansion, 1),
            ReLU(),
            Conv1d(d_model * expansion, d_model, 1),
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Apply ``PositionWiseFeedForward`` to the input.

        Args:
            x (torch.Tensor):

        Returns:

        """
        x = x.transpose(1, 2)
        x = self.conv(x)
        x = x.transpose(1, 2)
        return x


class ResidualSequential(Sequential):
    """

    Examples:
        >>> model = ResidualSequential(
        >>>     Linear(10, 10),
        >>>     ReLU(),
        >>>     Linear(10, 10)
        >>> )

    """

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward propagation

        Args:
            x: input data

        Returns:

        """
        return x + super(ResidualSequential, self).forward(x)


class AutoEncoder(Module):
    """
    A class that creates an ``AutoEncoder``.

    Examples:
        >>> ae = AutoEncoder([32, 16, 8])

    """

    def __init__(self, shapes: List[int], activation: Union[Callable[[Tensor], Tensor], Module] = relu):
        super(AutoEncoder, self).__init__()
        self.encoder = MLP(shapes, activation)
        self.decoder = MLP(list(reversed(shapes)), activation)

    def forward(self, x: Tensor) -> Tensor:
        """
        forward propagation processing.

        Args:
            x: input data

        Returns:

        """
        mid = self.encoder(x)
        out = self.decoder(mid)
        return out



from torch.tensor import Tensor
from torch import nn


__all__ = ["SELayer1d", "SELayer2d"]


class SELayer1d(nn.Module):
    """
    Squeeze and Excitation block for time series data.

    Paper: `Squeeze-and-Excitation Networks <https://arxiv.org/abs/1709.01507>`_
    """

    def __init__(self, in_features: int, reduction: int = 16) -> None:
        super(SELayer1d, self).__init__()
        reduction_size = max(1, in_features // reduction)

        self.avg_pool = nn.AdaptiveAvgPool1d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_features, reduction_size, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(reduction_size, in_features, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass

        Args:
            x: input data [N, C, L]

        Returns:

        """
        b, c, _ = x.shape
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1)
        scaling = x * y.expand_as(x)

        return scaling


class SELayer2d(nn.Module):
    """
    Squeeze and Excitation block for image data.

    Paper: `Squeeze-and-Excitation Networks <https://arxiv.org/abs/1709.01507>`_

    """

    def __init__(self, in_features: int, reduction: int = 16) -> None:
        super(SELayer2d, self).__init__()
        reduction_size = max(1, in_features // reduction)

        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_features, reduction_size, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(reduction_size, in_features, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass

        Args:
            x: input data [N, C, H, W]

        Returns:

        """
        b, c, _, _ = x.shape
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        scaling = x * y.expand_as(x)

        return scaling



from typing import Union
from sklearn.linear_model import LinearRegression
from sklearn.tree import BaseDecisionTree
from sklearn.svm._base import BaseLibSVM
from sklearn.neighbors._base import NeighborsBase

__all__ = ["ScikitModel"]


ScikitModel = Union[Union[LinearRegression, BaseDecisionTree], Union[BaseLibSVM, NeighborsBase]]



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import re
from typing import Any, Dict, Optional, Union

import torch
import numpy as np
from sklearn.svm import SVC
from sklearn.base import BaseEstimator
from sklearn.model_selection import cross_validate, StratifiedKFold, BaseCrossValidator, GridSearchCV, train_test_split
from .base import Callback
from ..engine.modules import fetch_state_dict


__all__ = ["EarlyStopping", "EarlyStoppingForTSUS"]


class EarlyStopping(Callback):
    """
    The training ends when the value to be monitored stops changing.

    Examples:
        >>> from enchanter.tasks import ClassificationRunner
        >>> runner = ClassificationRunner(callbacks=[ClassificationRunner()])

    """

    def __init__(
        self,
        monitor: str = "val_avg_loss",
        min_delta: float = 0.0,
        patience: int = 0,
        mode: str = "auto",
    ) -> None:
        """
        Initializer

        Args:
            monitor: You can choose the return values of ``train_end()``, ``val_end()``, or ``test_end()``.
                     To specify, in the case of the return value of ``train_end()``, ``train_XXX``,
                     in the case of the return value of ``val_end()``, ``val_XXX``,
                     in the case of the return value of ``test_end()``, ``test_XXX``.

            min_delta: Minimum value of change determined as improvement for the monitored value.
            patience: If there is no improvement in the monitored value during the specified number of epochs,
                      the training stops.
            mode: One of ``{'auto', 'min', 'max'}`` is selected.

                - ``min`` mode ends the training when the decrease in the monitored value stops.
                - ``max`` mode, the training is terminated when the monitored values stop increasing.
                - ``auto`` mode, automatically estimated from the monitored values.

        """
        super(EarlyStopping, self).__init__()
        if re.match("^validate_", monitor):
            monitor = monitor.replace("validate_", "val_")

        self.monitor: str = monitor
        self.patience: int = patience
        self.min_delta: float = min_delta
        self.wait: int = 0
        self.stopped_epoch: int = 0

        mode_dict = {
            "min": np.less,
            "max": np.greater,
            "auto": np.greater if "acc" in self.monitor else np.less,
        }
        if mode not in mode_dict:
            mode = "auto"

        self.monitor_op: Any = mode_dict[mode]
        self.min_delta *= 1 if self.monitor_op == np.greater else -1
        self.best: Any = np.Inf if self.monitor_op == np.less else -np.Inf

    def check_metrics(self, logs: Dict) -> bool:
        monitor_val = logs.get(self.monitor)

        if monitor_val is None:
            return False

        return True

    def on_epoch_end(self, epoch, logs: Optional[Dict] = None) -> None:
        stop = False

        if not isinstance(logs, dict):
            raise TypeError("The argument `logs` is not the expected data type.")
        else:
            cat_logs: Dict[str, Union[float, int]] = {}
            for pk in logs.keys():
                for ck, cv in logs[pk].items():
                    cat_logs["{}_{}".format(pk, ck)] = cv

            if not self.check_metrics(cat_logs):
                self.stop_runner = stop
            else:
                try:
                    current = cat_logs[self.monitor]
                except KeyError:
                    raise KeyError("Can't find the value specified in the argument `monitor`.")

                if self.monitor_op(current - self.min_delta, self.best):
                    self.best = current
                    self.wait = 0
                    self.params["best_weight"] = fetch_state_dict(self.model)

                else:
                    self.wait += 1
                    if self.wait > self.patience:
                        self.stopped_epoch = epoch
                        stop = True

                self.stop_runner = stop


class EarlyStoppingForTSUS(Callback):
    """
    Early Stopping for Time Series Unsupervised Runner.

    Examples:
        >>> import torch
        >>> from enchanter.tasks import TimeSeriesUnsupervisedRunner
        >>> x_train = torch.randn(32, 3, 128)
        >>> y_train = torch.randint(0, high=4, size=(128, ))
        >>> runner = TimeSeriesUnsupervisedRunner(
        >>>     callbacks=[EarlyStoppingForTSUS(x_train, y_train)]
        >>> )

    """

    def __init__(
        self,
        data: torch.Tensor,
        target: torch.Tensor,
        classifier: BaseEstimator = SVC(),
        monitor: str = "accuracy",
        min_delta: float = 0.0,
        patience: int = 0,
        kfold: Optional[BaseCrossValidator] = None,
        mode: str = "auto",
        grid_search: Optional[Dict[str, Any]] = None,
    ):
        """
        Initializer

        Args:
            data: Data for training a classifier to evaluate the quality of the encoder's output representation.
            target: Targets for training a classifier to evaluate the quality of the encoder's output representation.
            classifier: A classifier for evaluating the quality of the output representation of the encoder.
            monitor:
            min_delta:
            patience:
            kfold:
            mode:

        """
        super(EarlyStoppingForTSUS, self).__init__()
        self.classifier = classifier
        self.data: torch.Tensor = data
        self.target: torch.Tensor = target
        self.monitor: str = monitor
        self.patience = patience
        self.min_delta = min_delta
        self.best: float = 0.0
        self.wait: int = 0
        self.stopped_epoch: int = 0
        self.grid_search = grid_search

        if isinstance(kfold, BaseCrossValidator):
            self.kfold = kfold
        else:
            self.kfold = StratifiedKFold(n_splits=5, shuffle=True, random_state=0)

        mode_dict = {
            "min": np.less,
            "max": np.greater,
            "auto": np.greater if "acc" in self.monitor else np.less,
        }
        if mode not in mode_dict:
            mode = "auto"
        self.monitor_op: Any = mode_dict[mode]
        self.min_delta *= 1 if self.monitor_op == np.greater else -1

    def encode(self):
        out = []
        self.model.eval()
        with torch.no_grad():
            self.data = self.data.to(self.device)
            out.append(self.model(self.data).cpu().numpy())

        return np.vstack(out)

    def cross_val(self):
        features: np.ndarray = self.encode()
        targets = self.target.cpu().numpy()

        if self.grid_search is not None:
            evaluator = GridSearchCV(self.classifier, self.grid_search, cv=5, n_jobs=5)
            if len(targets) >= 10000:
                split = train_test_split(features, targets, train_size=10000, random_state=0, stratify=targets)
                features, targets = split[0], split[2]

            evaluator.fit(features, targets)
            score = evaluator.best_score_
            params = evaluator.best_params_

        else:
            evaluator = self.classifier

            score = cross_validate(
                evaluator,
                features,
                targets,
                cv=self.kfold,
                scoring=[self.monitor],
                n_jobs=self.kfold.n_splits,
            )["test_{}".format(self.monitor)].mean()
            params = evaluator.get_params()

        return score, params

    def on_epoch_end(self, epoch, logs: Optional[Dict] = None):
        stop = False
        current_score, current_params = self.cross_val()
        if self.experiment is not None:
            self.experiment.log_metric("Early Stopping metric", current_score, step=epoch)

        if self.monitor_op(current_score - self.min_delta, self.best):
            self.params["grid_search"] = current_params
            self.best = current_score
            self.wait = 0
            self.params["best_weight"] = fetch_state_dict(self.model)

        else:
            if self.wait > self.patience:
                self.stopped_epoch = epoch
                stop = True
            self.wait += 1

        self.stop_runner = stop
